# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function

import functools
import importlib
import os
import sys

import pytest
from packaging.version import Version

import pythonfinder

from .testutils import (
    is_in_ospath,
    normalize_path,
    normalized_match,
    print_python_versions,
)


@pytest.mark.skipif(sys.version_info < (3,), reason="Must run on Python 3")
def test_python_versions(monkeypatch, special_character_python):
    def mock_version(*args, **kwargs):
        version_output = "2.7.15+ (default, Jun 28 2018, 13:15:42)\n[GCC 7.2.0]"

        class FakeObj(object):
            def __init__(self, out):
                self.out = out

            def communicate(self):
                return self.out, ""

            def kill(self):
                pass

        c = FakeObj(version_output.split()[0])
        return c

    os.environ["PYTHONFINDER_IGNORE_UNSUPPORTED"] = str("1")
    with monkeypatch.context() as m:
        m.setattr("subprocess.Popen", mock_version)
        parsed = pythonfinder.models.python.PythonVersion.from_path(
            special_character_python.as_posix()
        )
        assert isinstance(parsed.version, Version)


@pytest.mark.parametrize(
    "path, version_output, version",
    [
        (
            "/fake/path/3.6.2/bin/python",
            "3.6.2 (default, Jun 28 2018, 13:15:42)",
            "3.6.2",
        ),
        (
            "/fake/path/3.7.0/bin/python",
            "3.7.0 (default, Jun 28 2018, 13:15:42)",
            "3.7.0",
        ),
        ("/fake/path/3.6.20/bin/python", "3.6.20 :: Continuum Analytics, Inc.", "3.6.20"),
        (
            "/fake/path/3.5.3/bin/python",
            "3.5.3 (fdd60ed87e941677e8ea11acf9f1819466521bf2, Apr 27 2018, 15:39:57)\n[PyPy 5.10.1 with GCC 4.2.1 Compatible Apple LLVM 9.0.0 (clang-900.0.39.2)]",
            "3.5.3",
        ),
        (
            "/fake/path/2.7.15+/bin/python",
            "2.7.15+ (default, Jun 28 2018, 13:15:42)",
            "2.7.15",
        ),
    ],
)
def test_python_version_output_variants(monkeypatch, path, version_output, version):
    def mock_version(*args, **kwargs):
        class FakeObj(object):
            def __init__(self, out):
                self.out = out

            def communicate(self):
                return self.out, ""

            def kill(self):
                pass

        c = FakeObj(".".join([str(i) for i in version_output.split()[0].split(".")]))
        return c

    def get_python_version(path, orig_fn=None):
        if not os.path.exists(path):
            return mock_version(version_output)
        return orig_fn(path)

    with monkeypatch.context() as m:
        os.environ["PYTHONFINDER_IGNORE_UNSUPPORTED"] = str("1")
        m.setattr("subprocess.Popen", mock_version)
        orig_run_fn = pythonfinder.utils.get_python_version
        get_pyversion = functools.partial(get_python_version, orig_fn=orig_run_fn)
        m.setattr("pythonfinder.utils.get_python_version", get_pyversion)
        # m.setattr("pythonfinder.utils.get_python_version", mock_version)
        parsed = pythonfinder.models.python.PythonVersion.from_path(path)
        assert isinstance(parsed.version, Version)


@pytest.mark.skip_nt
def test_shims_are_kept(monkeypatch, no_pyenv_root_envvar, setup_pythons, no_virtual_env):
    with monkeypatch.context() as m:
        os.environ["PATH"] = "{0}:{1}".format(
            normalize_path("~/.pyenv/shims"), os.environ["PATH"]
        )
        f = pythonfinder.pythonfinder.Finder(
            global_search=True, system=False, ignore_unsupported=True
        )
        f.rehash()
        # assert pythonfinder.environment.get_shim_paths() == []
        assert is_in_ospath("~/.pyenv/shims")
        shim_paths = pythonfinder.environment.get_shim_paths()
        # Shims directories are no longer added to the system path order
        # but instead are used as indicators of the presence of the plugin
        # and used to trigger plugin setup -- this is true only if ``PYENV_ROOT`` is set`
        if shim_paths:
            assert (
                os.path.join(normalize_path("~/.pyenv/shims"))
                not in f.system_path.path_order
            ), (
                pythonfinder.environment.get_shim_paths()
            )  # "\n".join(f.system_path.path_order)
        else:
            assert (
                os.path.join(normalize_path("~/.pyenv/shims")) in f.system_path.path_order
            ), "\n".join(f.system_path.path_order)
        python_versions = f.find_all_python_versions()
        anaconda = f.find_python_version("anaconda3-5.3.0")
        assert anaconda is not None, python_versions
        assert "shims" in anaconda.path.as_posix(), [
            f.system_path.path_order,
            f.system_path.pyenv_finder.roots,
        ]
        # for docker, use just 'anaconda'
        for path in f.system_path.path_order:
            print(
                "path: {0}    Entry: {1}".format(path, f.system_path.get_path(path)),
                file=sys.stderr,
            )
        which_anaconda = f.which("anaconda3-5.3.0")
        if shim_paths:
            assert "shims" not in which_anaconda.path.as_posix()
        else:
            assert "shims" in which_anaconda.path.as_posix()


@pytest.mark.skip_nt
def test_shims_are_removed(monkeypatch, no_virtual_env, setup_pythons):
    with monkeypatch.context() as m:
        pyenv_dir = pythonfinder.utils.normalize_path("~/.pyenv")
        asdf_dir = pythonfinder.utils.normalize_path("~/.asdf")
        importlib.reload(pythonfinder.environment)
        importlib.reload(pythonfinder.models.path)
        m.setattr(
            pythonfinder.environment,
            "SHIM_PATHS",
            pythonfinder.environment.get_shim_paths(),
        )
        f = pythonfinder.pythonfinder.Finder(
            global_search=True, system=False, ignore_unsupported=True
        )
        f.rehash()
        python_versions = f.find_all_python_versions()
        assert os.environ["PYENV_ROOT"] == os.path.abspath(
            os.path.join(os.path.expanduser("~"), ".pyenv")
        )
        assert os.environ["PYENV_ROOT"] == pythonfinder.environment.PYENV_ROOT
        assert pythonfinder.environment.PYENV_INSTALLED
        assert f.system_path.pyenv_finder is not None
        python_version_paths = list(
            v.path
            for v in python_versions
            if normalized_match(str(v.path), os.environ["PYENV_ROOT"])
        )
        # Make sure we have an entry for every python version installed
        python_names = set(list(v.parent.parent.name for v in python_version_paths))
        # this is how we test in docker / on local machines with python installs
        # setup_pythons = {version.name for version in all_python_versions if any(plugin in version.path for plugin in ("pyenv", "asdf"))}
        # this is the test implementation when we are simulating
        setup_key_list = [
            set(list(finder.keys())) for finder in list(setup_pythons.values())
        ]
        setup_pythons = {python for finder in setup_key_list for python in finder}
        # this calculates the pythons not present when we ran `find_all_python_versions`
        missing_from_finder = python_names ^ setup_pythons
        if missing_from_finder:
            print_python_versions(python_versions)
            for p in python_version_paths:
                print("path: {0}".format(p), file=sys.stderr)
            for p in sorted(python_names):
                print("python_name: {0}".format(p), file=sys.stderr)
            for p in sorted(list(setup_pythons)):
                print("setup python key: {0}".format(p), file=sys.stderr)
        assert not missing_from_finder, missing_from_finder
        anaconda = f.find_python_version("anaconda3-5.3.0")
        assert anaconda is not None, os.listdir(
            os.path.join(pyenv_dir, "versions", "anaconda3-5.3.0", "bin")
        )
        assert "shims" not in anaconda.path.as_posix()
        which_anaconda = f.which("anaconda3-5.3.0")
        # for docker, use just 'anaconda'
        # which_anaconda = f.which("anaconda")
        assert "shims" not in which_anaconda.path.as_posix()


@pytest.mark.skip_nt
def test_pythonfinder(expected_python_versions, all_python_versions):
    assert sorted(expected_python_versions) == sorted(all_python_versions)



# -*- coding=utf-8 -*-
from __future__ import absolute_import, print_function

import importlib
import os
import shutil
import subprocess
import sys
from collections import namedtuple

import click
import click.testing
import pytest

import pythonfinder

from .testutils import (
    cd,
    create_tracked_tempdir,
    normalize_path,
    normalized_match,
    set_write_bit,
    temp_environ,
    yield_versions,
)

if sys.version_info[:2] < (3, 5):
    from pathlib2 import Path
else:
    from pathlib import Path


pythoninfo = namedtuple("PythonVersion", ["name", "version", "path", "arch"])


def pytest_runtest_setup(item):
    if item.get_closest_marker("skip_nt") is not None and os.name == "nt":
        pytest.skip("does not run on windows")


@pytest.fixture
def pathlib_tmpdir(request, tmpdir):
    yield Path(str(tmpdir))
    try:
        tmpdir.remove(ignore_errors=True)
    except Exception:
        pass


def _create_tracked_dir():
    temp_args = {"prefix": "pythonfinder-", "suffix": "-test"}
    temp_args["dir"] = os.environ.get("TMPDIR", "/tmp")
    if temp_args["dir"] == "/":
        temp_args["dir"] = os.getcwd()
    temp_path = create_tracked_tempdir(**temp_args)
    return temp_path


@pytest.fixture
def tracked_tmpdir():
    temp_path = _create_tracked_dir()
    yield Path(temp_path)


@pytest.fixture(name="create_tmpdir")
def tracked_tmpdir_factory():
    def create_tmpdir():
        return Path(_create_tracked_dir())

    yield create_tmpdir


@pytest.fixture
def no_virtual_env():
    with temp_environ():
        if "VIRTUAL_ENV" in os.environ["PATH"]:
            orig_path = os.environ["PATH"].split(os.pathsep)
            venv = normalize_path(os.environ["VIRTUAL_ENV"])
            match = next(iter(p for p in orig_path if normalized_match(p, venv)), None)
            if match:
                orig_path.remove(match)
            del os.environ["VIRTUAL_ENV"]
            os.environ["PATH"] = os.pathsep.join(orig_path)
        yield


@pytest.fixture
def no_pyenv_root_envvar(monkeypatch):
    """This captures a bug hynek ran into where he had no `PYENV_ROOT`.

    When `PYENV_ROOT` is unset we still need to find pythons in the appropriate location
    """

    with monkeypatch.context() as m:
        if "PYENV_ROOT" in os.environ:
            m.delenv("PYENV_ROOT")
        if "ASDF_DATA_DIR" in os.environ:
            m.delenv("ASDF_DATA_DIR")
        importlib.reload(pythonfinder.environment)
        importlib.reload(pythonfinder.models.path)
        m.setattr(pythonfinder.environment, "PYENV_INSTALLED", False)
        m.setattr(pythonfinder.environment, "ASDF_INSTALLED", False)
        m.setattr(pythonfinder.environment, "PYENV_ROOT", normalize_path("~/.pyenv"))
        m.setattr(
            pythonfinder.environment,
            "ASDF_DATA_DIR",
            normalize_path("~/.asdf"),
        )
        m.setattr(
            pythonfinder.environment,
            "SHIM_PATHS",
            pythonfinder.environment.get_shim_paths(),
        )
        yield


@pytest.fixture
def isolated_envdir(create_tmpdir):
    runner = click.testing.CliRunner()
    fake_root_path = create_tmpdir()
    fake_root = fake_root_path.as_posix()
    set_write_bit(fake_root)
    with temp_environ(), cd(fake_root):
        home_dir_path = fake_root_path.joinpath("home/pythonfinder")
        home_dir_path.mkdir(parents=True)
        home_dir = normalize_path(home_dir_path.as_posix())
        os.chdir(home_dir)
        # This is pip's isolation approach, swipe it for now for time savings
        if sys.platform == "win32":
            home_drive, home_path = os.path.splitdrive(home_dir)
            os.environ.update(
                {"USERPROFILE": home_dir, "HOMEDRIVE": home_drive, "HOMEPATH": home_path}
            )
            for env_var, sub_path in (
                ("APPDATA", "AppData/Roaming"),
                ("LOCALAPPDATA", "AppData/Local"),
            ):
                path = os.path.join(home_dir, *sub_path.split("/"))
                os.environ[env_var] = path
                Path(path).mkdir(exist_ok=True, parents=True)
        else:
            os.environ["HOME"] = home_dir
            os.environ["XDG_DATA_HOME"] = os.path.join(home_dir, ".local", "share")
            os.environ["XDG_CONFIG_HOME"] = os.path.join(home_dir, ".config")
            os.environ["XDG_CACHE_HOME"] = os.path.join(home_dir, ".cache")
            os.environ["XDG_RUNTIME_DIR"] = os.path.join(home_dir, ".runtime")
            Path(os.path.join(home_dir, ".cache")).mkdir(exist_ok=True, parents=True)
            Path(os.path.join(home_dir, ".config")).mkdir(exist_ok=True, parents=True)
            Path(os.path.join(home_dir, ".local", "share")).mkdir(
                exist_ok=True, parents=True
            )
            Path(os.path.join(fake_root, "usr", "local", "share")).mkdir(
                exist_ok=True, parents=True
            )
            Path(os.path.join(fake_root, "usr", "share")).mkdir(
                exist_ok=True, parents=True
            )
            os.environ["XDG_DATA_DIRS"] = ":".join(
                [
                    os.path.join(fake_root, "usr", "local", "share"),
                    os.path.join(fake_root, "usr", "share"),
                ]
            )
        os.environ["PATH"] = os.defpath
        yield home_dir_path


def setup_plugin(name):
    target = os.path.expandvars(os.path.expanduser("~/.{0}".format(name)))
    this = Path(__file__).absolute().parent
    plugin_dir = this / "test_artifacts" / name
    plugin_uri = plugin_dir.as_uri()
    if not "file:///" in plugin_uri and "file:/" in plugin_uri:
        plugin_uri = plugin_uri.replace("file:/", "file:///")
    out = subprocess.check_output(["git", "clone", plugin_uri, Path(target).as_posix()])
    print(out, file=sys.stderr)


def build_python_versions(path, link_to=None):
    all_versions = {}
    for python_name, python_version in yield_versions():
        python_dir = path / python_name
        bin_dir = python_dir / "bin"
        bin_dir.mkdir(parents=True)
        set_write_bit(bin_dir.as_posix())
        executable_names = [
            "python",
            "python{0}".format(python_version[0]),
            "python{0}".format(python_version[:3]),
        ]
        for executable in executable_names:
            exe_file = bin_dir.joinpath(executable)
            shutil.copy2(sys.executable, str(exe_file))
        all_versions[python_name] = bin_dir / executable_names[-1]
        if link_to:
            target = link_to.joinpath(python_name)
            other_target = link_to.joinpath(executable_names[-1])
            if os.name == "nt":
                os.link(exe_file.as_posix(), target.as_posix())
                os.link(exe_file.as_posix(), other_target.as_posix())
            else:
                target.symlink_to(exe_file.as_posix())
                bin_target = bin_dir.joinpath(python_name)
                if not bin_target.exists():
                    bin_target.symlink_to(exe_file.as_posix())
                if not other_target.exists():
                    other_target.symlink_to(exe_file.as_posix())
                if python_name.startswith("anaconda"):
                    anaconda_target = link_to.joinpath("anaconda")
                    if not anaconda_target.exists():
                        anaconda_target.symlink_to(exe_file.as_posix())
    return all_versions


def setup_pyenv(home_dir):
    pyenv_dir = home_dir / ".pyenv"
    pyenv_shim_dir = pyenv_dir / "shims"
    pyenv_shim_dir.mkdir(parents=True)
    env_path = os.pathsep.join(
        [pyenv_shim_dir.as_posix(), os.environ.get("PATH", os.defpath)]
    )
    os.environ["PATH"] = env_path
    pyenv_python_dir = pyenv_dir / "versions"
    all_versions = build_python_versions(pyenv_python_dir, link_to=pyenv_shim_dir)
    os.environ["PYENV_ROOT"] = pyenv_dir.as_posix()
    return all_versions


def setup_asdf(home_dir):
    asdf_dir = home_dir / ".asdf"
    asdf_shim_dir = asdf_dir / "shims"
    asdf_shim_dir.mkdir(parents=True)
    env_path = os.pathsep.join(
        [asdf_shim_dir.as_posix(), os.environ.get("PATH", os.defpath)]
    )
    os.environ["PATH"] = env_path
    asdf_python_dir = asdf_dir / "installs/python"
    all_versions = build_python_versions(asdf_python_dir, link_to=asdf_shim_dir)
    os.environ["ASDF_DIR"] = asdf_dir.as_posix()
    os.environ["ASDF_DATA_DIR"] = asdf_dir.as_posix()
    return all_versions


@pytest.fixture
def setup_pythons(isolated_envdir, monkeypatch):
    with monkeypatch.context() as m:
        setup_plugin("asdf")
        setup_plugin("pyenv")
        asdf_dict = setup_asdf(isolated_envdir)
        pyenv_dict = setup_pyenv(isolated_envdir)
        os.environ["PATH"] = os.environ.get("PATH").replace("::", ":")
        version_dicts = {"pyenv": pyenv_dict, "asdf": asdf_dict}
        shim_paths = [
            normalize_path(isolated_envdir.joinpath(p).as_posix())
            for p in [".asdf/shims", ".pyenv/shims"]
        ]
        yield version_dicts


@pytest.fixture
def special_character_python(tmp_path):
    finder = pythonfinder.Finder(
        global_search=False, system=True, ignore_unsupported=True, sort_by_path=True
    )
    python = finder.find_python_version()
    python_name = "2+"
    python_folder = tmp_path / python_name / "bin"
    python_folder.mkdir(parents=True)
    set_write_bit(str(python_folder))
    python_path = python_folder / "python"
    python_path.symlink_to(python.path)
    return python_path


@pytest.fixture(autouse=True)
def setup_env():
    with temp_environ():
        os.environ["ANSI_COLORS_DISABLED"] = str("1")


@pytest.fixture()
def expected_python_versions():
    if os.name != "nt":
        return _build_python_tuples()
    return get_windows_python_versions()


def _build_python_tuples():
    finder = pythonfinder.Finder(
        global_search=True, system=False, ignore_unsupported=True
    )
    finder_versions = finder.find_all_python_versions()
    versions = []
    for v in finder_versions:
        if not v.is_python:
            continue
        version = v.as_python
        if not version:
            continue
        arch = (
            version.architecture
            if version.architecture
            else pythonfinder.environment.SYSTEM_ARCH
        )
        arch = arch.replace("bit", "")
        if version.name.startswith("python"):
            name = str(version.version)
        else:
            name = version.name
        path = normalize_path(v.path)
        version = str(version.version)
        versions.append(pythoninfo(name, version, path, arch))
    return versions


@pytest.fixture(scope="session")
def all_python_versions():
    return _build_python_tuples()


def get_windows_python_versions():
    out = subprocess.check_output("py -0p", shell=True)
    versions = []
    for line in out.splitlines():
        line = line.strip()
        if line and not "Installed Pythons found" in line:
            version, path = line.split("\t")
            version = version.strip().lstrip("-")
            path = normalize_path(path.strip().strip('"'))
            arch = None
            if "-" in version:
                version, _, arch = version.partition("-")
            versions.append(pythoninfo(version, version, path, arch))
    return versions



# -*- coding=utf-8 -*-

import os
import sys
from collections import namedtuple

import pytest

import pythonfinder.utils
from pythonfinder import Finder

from pathlib import Path

os.environ["ANSI_COLORS_DISABLED"] = "1"

pythoninfo = namedtuple("PythonVersion", ["version", "path", "arch"])


def _get_python_versions():
    finder = Finder(global_search=True, system=False, ignore_unsupported=True)
    pythons = finder.find_all_python_versions()
    for v in pythons:
        py = v.py_version
        comes_from = getattr(py, "comes_from", None)
        if comes_from is not None:
            comes_from_path = getattr(comes_from, "path", v.path)
        else:
            comes_from_path = v.path
    return sorted(list(pythons))


PYTHON_VERSIONS = _get_python_versions()


versions = [
    (
        pythonfinder.utils.get_python_version(python.path.as_posix()),
        python.as_python.version,
    )
    for python in PYTHON_VERSIONS
]

version_dicts = [
    (
        pythonfinder.utils.parse_python_version(str(python.as_python.version)),
        python.as_python.as_dict(),
    )
    for python in PYTHON_VERSIONS
]

test_paths = [(python.path.as_posix(), True) for python in PYTHON_VERSIONS]


@pytest.mark.parse
@pytest.mark.parametrize("python, expected", versions)
def test_get_version(python, expected):
    assert python == str(expected)


@pytest.mark.parse
@pytest.mark.parametrize("python, expected", version_dicts)
def test_parse_python_version(python, expected):
    for key in ["comes_from", "architecture", "executable", "name", "company"]:
        if key in expected:
            del expected[key]
        if key in python:
            del python[key]
    assert python == expected


@pytest.mark.is_python
@pytest.mark.parametrize("python, expected", test_paths)
def test_is_python(python, expected):
    assert pythonfinder.utils.path_is_known_executable(python)
    assert pythonfinder.utils.looks_like_python(os.path.basename(python))
    assert pythonfinder.utils.path_is_python(Path(python))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function

import atexit
import itertools
import os
import shutil
import stat
import sys
import tempfile
import warnings
from contextlib import contextmanager

import click

from pathlib import Path


TRACKED_TEMPORARY_DIRECTORIES = []


def set_write_bit(fn):
    # type: (str) -> None
    """
    Set read-write permissions for the current user on the target path.  Fail silently
    if the path doesn't exist.

    :param str fn: The target filename or path
    :return: None
    """

    if not os.path.exists(fn):
        return
    file_stat = os.stat(fn).st_mode
    os.chmod(fn, file_stat | stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
    if not os.path.isdir(fn):
        for path in [fn, os.path.dirname(fn)]:
            try:
                os.chflags(path, 0)
            except AttributeError:
                pass
        return None
    for root, dirs, files in os.walk(fn, topdown=False):
        for dir_ in [os.path.join(root, d) for d in dirs]:
            set_write_bit(dir_)
        for file_ in [os.path.join(root, f) for f in files]:
            set_write_bit(file_)


def create_tracked_tempdir(*args, **kwargs):
    """Create a tracked temporary directory.

    This uses ``tempfile.mkdtemp``, but does not remove the directory when
    the return value goes out of scope, instead registers a handler to cleanup
    on program exit.

    The return value is the path to the created directory.
    """

    tempdir = tempfile.mkdtemp(*args, **kwargs)
    TRACKED_TEMPORARY_DIRECTORIES.append(tempdir)
    atexit.register(shutil.rmtree, tempdir)
    warnings.simplefilter("ignore", ResourceWarning)
    return tempdir


@contextmanager
def cd(path):
    if not path:
        return
    prev_cwd = Path.cwd().as_posix()
    if isinstance(path, Path):
        path = path.as_posix()
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(prev_cwd)


@contextmanager
def temp_environ():
    """Allow the ability to set os.environ temporarily"""
    environ = dict(os.environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(environ)


def normalize_path(path):
    # type: (AnyStr) -> AnyStr
    """
    Return a case-normalized absolute variable-expanded path.

    :param str path: The non-normalized path
    :return: A normalized, expanded, case-normalized path
    :rtype: str
    """

    path = os.path.abspath(os.path.expandvars(os.path.expanduser(str(path))))
    if os.name == "nt" and os.path.exists(path):
        try:
            from ctypes import create_unicode_buffer, windll
        except ImportError:
            path = os.path.normpath(os.path.normcase(path))
        else:
            BUFSIZE = 500
            buffer = create_unicode_buffer(BUFSIZE)
            get_long_path_name = windll.kernel32.GetLongPathNameW
            get_long_path_name(path.decode(), buffer, BUFSIZE)
            path = buffer.value
        return path

    return os.path.normpath(os.path.normcase(path))


def is_in_path(path, parent):
    # type: (AnyStr, AnyStr) -> bool
    """
    Determine if the provided full path is in the given parent root.

    :param str path: The full path to check the location of.
    :param str parent: The parent path to check for membership in
    :return: Whether the full path is a member of the provided parent.
    :rtype: bool
    """

    return normalize_path(str(path)).startswith(normalize_path(str(parent)))


def normalized_match(path, parent):
    return is_in_path(normalize_path(path), normalize_path(parent))


def is_in_ospath(path):
    ospath = os.environ["PATH"].split(os.pathsep)
    return any(normalized_match(path, entry) for entry in ospath)


def print_python_versions(versions):
    versions = list(versions)
    if versions:
        click.echo("Found python at the following locations:", err=True)
        for v in versions:
            py = v.py_version
            comes_from = getattr(py, "comes_from", None)
            if comes_from is not None:
                comes_from_path = getattr(comes_from, "path", v.path)
            else:
                comes_from_path = v.path
            click.echo(
                "{py.name!s}: {py.version!s} ({py.architecture!s}) @ {comes_from!s}".format(
                    py=py, comes_from=comes_from_path
                ),
                err=True,
            )


STACKLESS = [("stackless-2.7.14", "2.7.14"), ("stackless-3.5.4", "3.5.4")]
PYPY = [
    ("pypy2.7-5.10.0", "2.7"),
    ("pypy2.7-6.0.0", "2.7"),
    ("pypy3.5-5.10.1", "3.5"),
    ("pypy3.5-6.0.0", "3.5"),
]
# These are definitely no longer supported but people may still have them installed
# PYSTON = [("pyston-0.5.1", "2.7.8"), ("pyston-0.6.0", "2.7.8"), ("pyston-0.6.1", "2.7.8")]
MINICONDA = [
    ("miniconda3-3.18.3", "3.5.4"),
    ("miniconda3-3.19.0", "3.5.5"),
    ("miniconda3-4.0.5", "3.6.2"),
    ("miniconda3-4.1.11", "3.6.2"),
    ("miniconda3-4.3.27", "3.7.1"),
    ("miniconda3-4.3.30", "3.7.2"),
]
# I have no idea how these are represented
# MICROPYTHON = [("micropython-1.9.3", "1.9.3"), ("micropython-1.9.4", "1.9.4")]
JYTHON = [
    # ("jython-2.5-dev", "2.5-dev"),
    # ("jython-2.5.1", "2.5.1"),
    # ("jython-2.5.2", "2.5.2"),
    # ("jython-2.5.3", "2.5.3"),
    # ("jython-2.5.4-rc1", "2.5.4-rc1"),
    # ("jython-2.7.0", "2.7.0"),
    ("jython-2.7.1", "2.7.1")
]
ANACONDA = [
    ("anaconda2-2.4.1", "2.7.8"),
    ("anaconda2-2.5.0", "2.7.10"),
    ("anaconda2-4.0.0", "2.7.12"),
    ("anaconda2-4.1.0", "2.7.13"),
    ("anaconda2-4.1.1", "2.7.13"),
    ("anaconda2-4.2.0", "2.7.14"),
    ("anaconda2-4.3.0", "2.7.14"),
    ("anaconda2-4.3.1", "2.7.15"),
    ("anaconda2-4.4.0", "2.7.15"),
    ("anaconda3-4.1.1", "3.4.5"),
    ("anaconda3-4.2.0", "3.5.2"),
    ("anaconda3-4.3.0", "3.5.3"),
    ("anaconda3-4.3.1", "3.5.4"),
    ("anaconda3-4.4.0", "3.5.5"),
    ("anaconda3-5.0.0", "3.6.2"),
    ("anaconda3-5.0.1", "3.6.2"),
    ("anaconda3-5.1.0", "3.6.3"),
    ("anaconda3-5.2.0", "3.6.4"),
    ("anaconda3-5.3.0", "3.6.5"),
    ("anaconda3-2019.03", "3.7.2"),
]
IRONPYTHON = [
    # ("ironpython-dev", "2.7.8"),
    # ("ironpython-2.7.4", "2.7.4"),
    # ("ironpython-2.7.5", "2.7.5"),
    # ("ironpython-2.7.6.3", "2.7.6"),
    ("ironpython-2.7.7", "2.7.7")
]
ACTIVEPYTHON = [
    ("activepython-2.7.14", "2.7.14"),
    ("activepython-3.5.4", "3.5.4"),
    ("activepython-3.6.0", "3.6.0"),
]
PYTHON = [
    ("2.7.14", "2.7.14"),
    ("2.7.15", "2.7.15"),
    ("3.6.6", "3.6.6"),
    ("3.6.7", "3.6.7"),
    ("3.6.8", "3.6.8"),
    ("3.7.0", "3.7.0"),
    ("3.7-dev", "3.7-dev"),
    ("3.7.1", "3.7.1"),
    ("3.7.2", "3.7.2"),
    ("3.7.3", "3.7.3"),
    ("3.8-dev", "3.8-dev"),
]


def yield_versions():
    return itertools.chain(
        STACKLESS,
        PYPY,
        # PYSTON,
        MINICONDA,
        # MICROPYTHON,
        JYTHON,
        ANACONDA,
        IRONPYTHON,
        ACTIVEPYTHON,
        PYTHON,
    )
